package apis

import (
	"fmt"

	"github.com/gin-gonic/gin"
	"github.com/go-admin-team/go-admin-core/sdk/api"
	"github.com/go-admin-team/go-admin-core/sdk/pkg/jwtauth/user"
	_ "github.com/go-admin-team/go-admin-core/sdk/pkg/response"

	"go-admin/app/admin/models"
	"go-admin/app/admin/service"
	"go-admin/app/admin/service/dto"
	"go-admin/common/actions"
)

type OtIfram struct {
	api.Api
}

// GetPage 获取OtIfram列表
// @Summary 获取OtIfram列表
// @Description 获取OtIfram列表
// @Tags OtIfram
// @Param url query string false "iframe地址"
// @Param pageTag query string false "页面标识"
// @Param appId query int64 false "AppID"
// @Param pageSize query int false "页条数"
// @Param pageIndex query int false "页码"
// @Success 200 {object} response.Response{data=response.Page{list=[]models.OtIfram}} "{"code": 200, "data": [...]}"
// @Router /api/v1/ot-ifram [get]
// @Security Bearer
func (e OtIfram) GetPage(c *gin.Context) {
	req := dto.OtIframGetPageReq{}
	s := service.OtIfram{}
	err := e.MakeContext(c).
		MakeOrm().
		Bind(&req).
		MakeService(&s.Service).
		Errors
	if err != nil {
		e.Logger.Error(err)
		e.Error(500, err, err.Error())
		return
	}

	p := actions.GetPermissionFromContext(c)
	list := make([]models.OtIfram, 0)
	var count int64

	err = s.GetPage(&req, p, &list, &count)
	if err != nil {
		e.Error(500, err, fmt.Sprintf("获取OtIfram失败，\r\n失败信息 %s", err.Error()))
		return
	}

	e.PageOK(list, int(count), req.GetPageIndex(), req.GetPageSize(), "查询成功")
}

// Get 获取OtIfram
// @Summary 获取OtIfram
// @Description 获取OtIfram
// @Tags OtIfram
// @Param id path int false "id"
// @Success 200 {object} response.Response{data=models.OtIfram} "{"code": 200, "data": [...]}"
// @Router /api/v1/ot-ifram/{id} [get]
// @Security Bearer
func (e OtIfram) Get(c *gin.Context) {
	req := dto.OtIframGetReq{}
	s := service.OtIfram{}
	err := e.MakeContext(c).
		MakeOrm().
		Bind(&req).
		MakeService(&s.Service).
		Errors
	if err != nil {
		e.Logger.Error(err)
		e.Error(500, err, err.Error())
		return
	}
	var object models.OtIfram

	p := actions.GetPermissionFromContext(c)
	err = s.Get(&req, p, &object)
	if err != nil {
		e.Error(500, err, fmt.Sprintf("获取OtIfram失败，\r\n失败信息 %s", err.Error()))
		return
	}

	e.OK(object, "查询成功")
}

// Insert 创建OtIfram
// @Summary 创建OtIfram
// @Description 创建OtIfram
// @Tags OtIfram
// @Accept application/json
// @Product application/json
// @Param data body dto.OtIframInsertReq true "data"
// @Success 200 {object} response.Response	"{"code": 200, "message": "添加成功"}"
// @Router /api/v1/ot-ifram [post]
// @Security Bearer
func (e OtIfram) Insert(c *gin.Context) {
	req := dto.OtIframInsertReq{}
	s := service.OtIfram{}
	err := e.MakeContext(c).
		MakeOrm().
		Bind(&req).
		MakeService(&s.Service).
		Errors
	if err != nil {
		e.Logger.Error(err)
		e.Error(500, err, err.Error())
		return
	}
	// 设置创建人
	req.SetCreateBy(user.GetUserId(c))

	err = s.Insert(&req)
	if err != nil {
		e.Error(500, err, fmt.Sprintf("创建OtIfram失败，\r\n失败信息 %s", err.Error()))
		return
	}

	e.OK(req.GetId(), "创建成功")
}

// Update 修改OtIfram
// @Summary 修改OtIfram
// @Description 修改OtIfram
// @Tags OtIfram
// @Accept application/json
// @Product application/json
// @Param id path int true "id"
// @Param data body dto.OtIframUpdateReq true "body"
// @Success 200 {object} response.Response	"{"code": 200, "message": "修改成功"}"
// @Router /api/v1/ot-ifram/{id} [put]
// @Security Bearer
func (e OtIfram) Update(c *gin.Context) {
	req := dto.OtIframUpdateReq{}
	s := service.OtIfram{}
	err := e.MakeContext(c).
		MakeOrm().
		Bind(&req).
		MakeService(&s.Service).
		Errors
	if err != nil {
		e.Logger.Error(err)
		e.Error(500, err, err.Error())
		return
	}
	req.SetUpdateBy(user.GetUserId(c))
	p := actions.GetPermissionFromContext(c)

	err = s.Update(&req, p)
	if err != nil {
		e.Error(500, err, fmt.Sprintf("修改OtIfram失败，\r\n失败信息 %s", err.Error()))
		return
	}
	e.OK(req.GetId(), "修改成功")
}

// Delete 删除OtIfram
// @Summary 删除OtIfram
// @Description 删除OtIfram
// @Tags OtIfram
// @Param data body dto.OtIframDeleteReq true "body"
// @Success 200 {object} response.Response	"{"code": 200, "message": "删除成功"}"
// @Router /api/v1/ot-ifram [delete]
// @Security Bearer
func (e OtIfram) Delete(c *gin.Context) {
	s := service.OtIfram{}
	req := dto.OtIframDeleteReq{}
	err := e.MakeContext(c).
		MakeOrm().
		Bind(&req).
		MakeService(&s.Service).
		Errors
	if err != nil {
		e.Logger.Error(err)
		e.Error(500, err, err.Error())
		return
	}

	// req.SetUpdateBy(user.GetUserId(c))
	p := actions.GetPermissionFromContext(c)

	err = s.Remove(&req, p)
	if err != nil {
		e.Error(500, err, fmt.Sprintf("删除OtIfram失败，\r\n失败信息 %s", err.Error()))
		return
	}
	e.OK(req.GetId(), "删除成功")
}
