package apis

import (
	"fmt"

	"github.com/gin-gonic/gin"
	"github.com/go-admin-team/go-admin-core/sdk/api"
	"github.com/go-admin-team/go-admin-core/sdk/pkg/jwtauth/user"
	_ "github.com/go-admin-team/go-admin-core/sdk/pkg/response"

	"go-admin/app/admin/models"
	"go-admin/app/admin/service"
	"go-admin/app/admin/service/dto"
	"go-admin/common/actions"
)

type OtPlugins struct {
	api.Api
}

// GetPage 获取OtPlugins列表
// @Summary 获取OtPlugins列表
// @Description 获取OtPlugins列表
// @Tags OtPlugins
// @Param name query string false "插件名"
// @Param alias query string false "插件英文别名"
// @Param pageSize query int false "页条数"
// @Param pageIndex query int false "页码"
// @Success 200 {object} response.Response{data=response.Page{list=[]models.OtPlugins}} "{"code": 200, "data": [...]}"
// @Router /api/v1/ot-plugins [get]
// @Security Bearer
func (e OtPlugins) GetPage(c *gin.Context) {
	req := dto.OtPluginsGetPageReq{}
	s := service.OtPlugins{}
	err := e.MakeContext(c).
		MakeOrm().
		Bind(&req).
		MakeService(&s.Service).
		Errors
	if err != nil {
		e.Logger.Error(err)
		e.Error(500, err, err.Error())
		return
	}

	p := actions.GetPermissionFromContext(c)
	list := make([]models.OtPlugins, 0)
	var count int64

	err = s.GetPage(&req, p, &list, &count)
	if err != nil {
		e.Error(500, err, fmt.Sprintf("获取OtPlugins失败，\r\n失败信息 %s", err.Error()))
		return
	}

	e.PageOK(list, int(count), req.GetPageIndex(), req.GetPageSize(), "查询成功")
}

// Get 获取OtPlugins
// @Summary 获取OtPlugins
// @Description 获取OtPlugins
// @Tags OtPlugins
// @Param id path int false "id"
// @Success 200 {object} response.Response{data=models.OtPlugins} "{"code": 200, "data": [...]}"
// @Router /api/v1/ot-plugins/{id} [get]
// @Security Bearer
func (e OtPlugins) Get(c *gin.Context) {
	req := dto.OtPluginsGetReq{}
	s := service.OtPlugins{}
	err := e.MakeContext(c).
		MakeOrm().
		Bind(&req).
		MakeService(&s.Service).
		Errors
	if err != nil {
		e.Logger.Error(err)
		e.Error(500, err, err.Error())
		return
	}
	var object models.OtPlugins

	p := actions.GetPermissionFromContext(c)
	err = s.Get(&req, p, &object)
	if err != nil {
		e.Error(500, err, fmt.Sprintf("获取OtPlugins失败，\r\n失败信息 %s", err.Error()))
		return
	}

	e.OK(object, "查询成功")
}

// Insert 创建OtPlugins
// @Summary 创建OtPlugins
// @Description 创建OtPlugins
// @Tags OtPlugins
// @Accept application/json
// @Product application/json
// @Param data body dto.OtPluginsInsertReq true "data"
// @Success 200 {object} response.Response	"{"code": 200, "message": "添加成功"}"
// @Router /api/v1/ot-plugins [post]
// @Security Bearer
func (e OtPlugins) Insert(c *gin.Context) {
	req := dto.OtPluginsInsertReq{}
	s := service.OtPlugins{}
	err := e.MakeContext(c).
		MakeOrm().
		Bind(&req).
		MakeService(&s.Service).
		Errors
	if err != nil {
		e.Logger.Error(err)
		e.Error(500, err, err.Error())
		return
	}
	// 设置创建人
	req.SetCreateBy(user.GetUserId(c))

	err = s.Insert(&req)
	if err != nil {
		e.Error(500, err, fmt.Sprintf("创建OtPlugins失败，\r\n失败信息 %s", err.Error()))
		return
	}

	e.OK(req.GetId(), "创建成功")
}

// Update 修改OtPlugins
// @Summary 修改OtPlugins
// @Description 修改OtPlugins
// @Tags OtPlugins
// @Accept application/json
// @Product application/json
// @Param id path int true "id"
// @Param data body dto.OtPluginsUpdateReq true "body"
// @Success 200 {object} response.Response	"{"code": 200, "message": "修改成功"}"
// @Router /api/v1/ot-plugins/{id} [put]
// @Security Bearer
func (e OtPlugins) Update(c *gin.Context) {
	req := dto.OtPluginsUpdateReq{}
	s := service.OtPlugins{}
	err := e.MakeContext(c).
		MakeOrm().
		Bind(&req).
		MakeService(&s.Service).
		Errors
	if err != nil {
		e.Logger.Error(err)
		e.Error(500, err, err.Error())
		return
	}
	req.SetUpdateBy(user.GetUserId(c))
	p := actions.GetPermissionFromContext(c)

	err = s.Update(&req, p)
	if err != nil {
		e.Error(500, err, fmt.Sprintf("修改OtPlugins失败，\r\n失败信息 %s", err.Error()))
		return
	}
	e.OK(req.GetId(), "修改成功")
}

// Delete 删除OtPlugins
// @Summary 删除OtPlugins
// @Description 删除OtPlugins
// @Tags OtPlugins
// @Param data body dto.OtPluginsDeleteReq true "body"
// @Success 200 {object} response.Response	"{"code": 200, "message": "删除成功"}"
// @Router /api/v1/ot-plugins [delete]
// @Security Bearer
func (e OtPlugins) Delete(c *gin.Context) {
	s := service.OtPlugins{}
	req := dto.OtPluginsDeleteReq{}
	err := e.MakeContext(c).
		MakeOrm().
		Bind(&req).
		MakeService(&s.Service).
		Errors
	if err != nil {
		e.Logger.Error(err)
		e.Error(500, err, err.Error())
		return
	}

	// req.SetUpdateBy(user.GetUserId(c))
	p := actions.GetPermissionFromContext(c)

	err = s.Remove(&req, p)
	if err != nil {
		e.Error(500, err, fmt.Sprintf("删除OtPlugins失败，\r\n失败信息 %s", err.Error()))
		return
	}
	e.OK(req.GetId(), "删除成功")
}

func (e OtPlugins) GetPluginsWithRole(c *gin.Context) {
	req := dto.OtPluginsGetPageReq{}
	s := service.OtPlugins{}
	err := e.MakeContext(c).
		MakeOrm().
		Bind(&req).
		MakeService(&s.Service).
		Errors
	if err != nil {
		e.Logger.Error(err)
		e.Error(500, err, err.Error())
		return
	}

	p := actions.GetPermissionFromContext(c)
	list := make([]models.OtPlugins, 0)
	var count int64

	err = s.GetPage(&req, p, &list, &count)
	if err != nil {
		e.Error(500, err, fmt.Sprintf("获取OtPlugins失败，\r\n失败信息 %s", err.Error()))
		return
	}

	e.PageOK(list, int(count), req.GetPageIndex(), req.GetPageSize(), "查询成功")
}
