/*
Copyright © 2023 NAME HERE <EMAIL ADDRESS>

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

	http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/
package gen

import (
	"bytes"
	"fmt"
	"log"
	"unicode"

	"go-admin/common/database"
	"go-admin/common/olap"
	extConfig "go-admin/config"

	"github.com/go-admin-team/go-admin-core/config/source/file"
	"github.com/go-admin-team/go-admin-core/sdk"
	"github.com/go-admin-team/go-admin-core/sdk/config"
	"github.com/spf13/cobra"
	"gorm.io/gen"
	"gorm.io/gorm"
	"gorm.io/gorm/schema"
)

// ch2mdlCmd represents the ch2mdl command
var StartCmd = &cobra.Command{
	Use:   "gen-model",
	Short: "clickhouse 表转化为 model",
	Long:  `通过该命令，可将clickhouse中的表转化为gorm中的model`,
	PreRun: func(cmd *cobra.Command, args []string) {
		// 注入配置扩展项
		config.ExtendConfig = &extConfig.ExtConfig
		//1. 读取配置
		config.Setup(
			file.NewSource(file.WithPath(configYml)),
			database.Setup,
			olap.Setup,
		)
	},
	RunE: func(cmd *cobra.Command, args []string) error {
		tables := args
		fmt.Println(extConfig.ExtConfig)
		// cmdParse
		config := argParse(args)
		if config == nil {
			log.Fatalln("parse config fail")
		}

		db, err := connectDB()
		if err != nil {
			log.Fatalln("connect db server fail:", err)
		}

		g := gen.NewGenerator(gen.Config{
			OutPath:           config.OutPath,
			OutFile:           config.OutFile,
			ModelPkgPath:      config.ModelPkgName,
			WithUnitTest:      config.WithUnitTest,
			FieldNullable:     config.FieldNullable,
			FieldCoverable:    config.FieldCoverable,
			FieldWithIndexTag: config.FieldWithIndexTag,
			FieldWithTypeTag:  config.FieldWithTypeTag,
			FieldSignable:     config.FieldSignable,
		})

		g.UseDB(db)
		g.Config.WithJSONTagNameStrategy(func(columnName string) string {
			var output bytes.Buffer
			for i, v := range columnName {
				if unicode.IsUpper(v) {
					if i > 0 {
						output.WriteRune('_')
					}
					output.WriteRune(unicode.ToLower(v))
				} else {
					output.WriteRune(v)
				}
			}
			return output.String()
		})

		models, err := genModels(g, db, tables)
		if err != nil {
			log.Fatalln("get tables info fail:", err)
		}

		if !config.OnlyModel {
			g.ApplyBasic(models...)
		}

		g.Execute()
		return nil
	},
}
var configYml string
var dbType string

func init() {
	StartCmd.PersistentFlags().StringVarP(&configYml, "config", "c", "config/settings.yml", "配置文件")
	StartCmd.PersistentFlags().StringVarP(&dbType, "dbtype", "t", "clickhouse", "数据库类型 mysql or clickhouse, 默认clickhouse")
}

// CmdParams is command line parameters
type CmdParams struct {
	DB                string   `yaml:"db"`                // input mysql or postgres or sqlite or sqlserver. consult[https://gorm.io/docs/connecting_to_the_database.html]
	Tables            []string `yaml:"tables"`            // enter the required data table or leave it blank
	OnlyModel         bool     `yaml:"onlyModel"`         // only generate model
	OutPath           string   `yaml:"outPath"`           // specify a directory for output
	OutFile           string   `yaml:"outFile"`           // query code file name, default: gen.go
	WithUnitTest      bool     `yaml:"withUnitTest"`      // generate unit test for query code
	ModelPkgName      string   `yaml:"modelPkgName"`      // generated model code's package name
	FieldNullable     bool     `yaml:"fieldNullable"`     // generate with pointer when field is nullable
	FieldCoverable    bool     `yaml:"fieldCoverable"`    // generate with pointer when field has default value
	FieldWithIndexTag bool     `yaml:"fieldWithIndexTag"` // generate field with gorm index tag
	FieldWithTypeTag  bool     `yaml:"fieldWithTypeTag"`  // generate field with gorm column type tag
	FieldSignable     bool     `yaml:"fieldSignable"`     // detect integer field's unsigned type, adjust generated data type
}

// connectDB choose db type for connection to database
func connectDB() (*gorm.DB, error) {
	switch dbType {
	case "clickhouse":
		db := olap.GetClickhouseOrm()
		db.Config.NamingStrategy = olap.ChNamingStrategy{
			NamingStrategy: schema.NamingStrategy{TablePrefix: "otel_"},
		}
		return db, nil
	case "mysql":
		host := "*"
		db := sdk.Runtime.GetDbByKey(host)
		db.Set("gorm:table_options", "ENGINE=InnoDB CHARSET=utf8mb4")
		db.Config.NamingStrategy = schema.NamingStrategy{
			TablePrefix: "ot_",
		}
		return db, nil
	default:
		return nil, fmt.Errorf("unknow db %q (support mysql || clickhouse for now)", dbType)
	}
}

// genModels is gorm/gen generated models
func genModels(g *gen.Generator, db *gorm.DB, tables []string) (models []interface{}, err error) {
	if len(tables) == 0 {
		// Execute tasks for all tables in the database
		tables, err = db.Migrator().GetTables()
		if err != nil {
			return nil, fmt.Errorf("GORM migrator get all tables fail: %w", err)
		}
	}

	// Execute some data table tasks
	models = make([]interface{}, len(tables))
	for i, tableName := range tables {
		models[i] = g.GenerateModel(tableName)
	}
	return models, nil
}

// argParse is parser for cmd
func argParse(tables []string) *CmdParams {
	return &CmdParams{
		OnlyModel:         true,
		OutPath:           "./app/observe/models/",
		WithUnitTest:      false,
		ModelPkgName:      "models",
		FieldNullable:     false,
		FieldCoverable:    false,
		FieldWithIndexTag: true,
		FieldWithTypeTag:  true,
		FieldSignable:     false,
	}
}
