package ws

import (
	"context"
	"encoding/json"
	"go-admin/app/observe/apis"
	"go-admin/app/observe/models"
	"go-admin/app/observe/service"
	"go-admin/app/observe/service/dto"
	"go-admin/render"
	"go-admin/render/report"
	"net/http"
	"time"

	"github.com/gin-gonic/gin"
	"github.com/go-admin-team/go-admin-core/logger"
	"github.com/gorilla/websocket"
	uuid "github.com/satori/go.uuid"
)

type ScopeGraphBasic struct {
	ADD    []render.NodeSummary `json:"add"`
	Update []render.NodeSummary `json:"update"`
	Remove []render.NodeSummary `json:"remove"`
	Reset  bool                 `json:"reset"`
}

func (manager *Manager) ServiceTopoGraphWSClient(ctx *gin.Context) {
	upGrader := websocket.Upgrader{
		CheckOrigin: func(r *http.Request) bool {
			return true
		},
		Subprotocols: []string{ctx.GetHeader("Sec-WebSocket-Protocol")},
	}

	conn, err := upGrader.Upgrade(ctx.Writer, ctx.Request, nil)
	if err != nil {
		logger.Debugf("websocket connect error: %s", ctx.Param("channel"))
		return
	}
	client := &Client{
		ID:      uuid.NewV4().String(),
		Group:   ctx.Param("channel"),
		Socket:  conn,
		Message: make(chan []byte, 1024),
	}
	logger.Debug("client ", client)
	manager.RegisterClient(client)
	defer func() {
		conn.Close()
		manager.UnRegisterClient(client)
		logger.Info("ws client closed")
	}()
	go client.Read()
	go client.Write()
	api := apis.Service{}
	req := new(dto.ServiceGetEdgesReq)
	svc := new(service.Service)
	if err := api.Init(ctx, req, &svc.OtService); err != nil {
		logger.Errorf("bind ws params err: %v", err)
		return
	}
	logger.Debugf("starttime: %v,endtime: %v, live: %v, req-before: %v", req.StartTime, req.EndTime, req.Live, req)
	if req.StartTime == 0 && req.EndTime == 0 { // 兼容前端
		req.EndTime = time.Now().Unix()
		req.StartTime = req.EndTime - 600 // 10分钟
		req.Live = true
	}
	if req.AppAlias == "" {
		req.AppAlias = "UNSET"
	}
	resp := new(models.Graph)
	//?t=5,默认使用5s
	add := true
	interval := req.EndTime - req.StartTime
	if interval > 60*3 {
		interval = 60 * 3
	}
	newTick := make(chan struct{})
	sendQueue := make(chan ScopeGraphBasic, 1)
	go func() {
		ticker := time.NewTicker(100 * time.Microsecond)
		defer ticker.Stop()
		for {
			select {
			case <-manager.UnRegister:
				logger.Debug("exist mission of tick 100ms")
				return
			case <-ticker.C:
				if len(sendQueue) < cap(sendQueue) {
					sgb := new(ScopeGraphBasic)
					if add {
						sgb.addNoSoulGraph(svc, ctx, req, resp)
						sendQueue <- *sgb
						newTick <- struct{}{}

						if req.Live {
							req.EndTime = time.Now().Unix()
							req.StartTime = req.EndTime - interval
						}
						s2 := new(ScopeGraphBasic)
						s2.UpdateGraph(svc, ctx, req, resp)
						sendQueue <- *s2
						newTick <- struct{}{}
					} else {
						if req.Live {
							req.EndTime = time.Now().Unix()
							req.StartTime = req.EndTime - interval
						}
						sgb.UpdateGraph(svc, ctx, req, resp)
						sendQueue <- *sgb
					}
					add = false
				}

			}
		}
	}()

	ticker := time.NewTicker(5 * time.Second)
	defer ticker.Stop()
	for {
		select {
		case <-manager.UnRegister:
			return
		case <-ticker.C:
			b, _ := json.Marshal(<-sendQueue)
			manager.Send(client.ID, client.Group, (b))
		case <-newTick:
			b, _ := json.Marshal(<-sendQueue)
			manager.Send(client.ID, client.Group, (b))
		}
	}
}

func (sgb *ScopeGraphBasic) setShape(icon string) string {
	switch icon {
	case "cloud":
		return "cloud"
	default:
		return "circle"
	}
}

func (sgb *ScopeGraphBasic) addNoSoulGraph(
	svc *service.Service,
	ctx context.Context,
	req *dto.ServiceGetEdgesReq,
	resp *models.Graph) {
	if err := svc.GetNoSoulGraph(ctx, req, resp); err != nil {
		logger.Error("get no-soul graph err: ", err.Error())
		// api.Error(http.StatusInternalServerError, err, err.Error())
		return
	}
	for _, node := range resp.Nodes {
		ns := render.NodeSummary{
			BasicNodeSummary: render.BasicNodeSummary{
				ID:         node.ID,
				Label:      node.Title,    //钻取下级  从属依赖下级
				LabelMinor: node.SubTitle, //钻取下下级  从属依赖下下级
				Shape:      sgb.setShape(node.Icon),
				Color:      "DI",
			},
		}
		//TODO:
		for _, es := range resp.Edges {
			if es.Source == node.ID {
				ns.Adjacency = append(ns.Adjacency, es.Target)
			}
		}
		sgb.ADD = append(sgb.ADD, ns)
	}
	sgb.Reset = true
}

func (sgb *ScopeGraphBasic) UpdateGraph(
	svc *service.Service,
	ctx context.Context,
	req *dto.ServiceGetEdgesReq,
	resp *models.Graph) {
	if err := svc.GetGraph(ctx, req, resp); err != nil {
		logger.Error("get graph err: ", err.Error())
		// api.Error(http.StatusInternalServerError, err, err.Error())
		return
	}
	for _, node := range resp.Nodes {
		ns := render.NodeSummary{
			BasicNodeSummary: render.BasicNodeSummary{
				ID:         node.ID,
				Label:      node.Title,    //钻取下级  从属依赖下级
				LabelMinor: node.SubTitle, //钻取下下级  从属依赖下下级
				Shape:      sgb.setShape(node.Icon),
				Color:      "G",
				// Tag:        "camera",
			},
		}
		mrApdex := report.MetricRow{
			ID:       "apdex",
			Label:    "Apdex",
			Format:   "percent",
			Value:    node.Apdex * 100,
			Priority: 1,
			Metric: report.Metric{
				Min: 0,
				Max: 100,
			},
		}
		mrErr := report.MetricRow{
			ID:       "request failed",
			Label:    "Error",
			Format:   "percent",
			Value:    node.ArcFaild * 100,
			Priority: 1,
			Metric: report.Metric{
				Min: 0,
				Max: 100,
			},
		}
		mrSucc := report.MetricRow{
			ID:       "request success",
			Label:    "Success",
			Format:   "percent",
			Value:    node.ArcSuccess * 100,
			Priority: 1,
			Metric: report.Metric{
				Min: 0,
				Max: 100,
			},
		}
		mrSend := report.MetricRow{
			ID:       "Send Req",
			Label:    "Send",
			Format:   "",
			Value:    float64(node.Send),
			Priority: 1,
			Metric: report.Metric{
				Min: 0,
				Max: 600,
			},
		}
		mrReceive := report.MetricRow{
			ID:       "Receive Req",
			Label:    "Receive",
			Format:   "",
			Value:    float64(node.Receive),
			Priority: 1,
			Metric: report.Metric{
				Min: 1,
				Max: 600,
			},
		}
		if node.ArcFaild+node.ArcSuccess == 0 {
			ns.Color = "DI"
		} else {
			if node.ArcSuccess < 0.6 || node.Apdex < 0.4 {
				ns.Color = "R"
			} else if (node.ArcSuccess >= 0.6 && node.ArcSuccess < 0.8) ||
				(node.Apdex >= 0.4 && node.Apdex < 0.8) {
				ns.Color = "Y"
			}
		}
		ns.Metrics = append(ns.Metrics, mrApdex)
		ns.Metrics = append(ns.Metrics, mrErr)
		ns.Metrics = append(ns.Metrics, mrSucc)
		ns.Metrics = append(ns.Metrics, mrSend)
		ns.Metrics = append(ns.Metrics, mrReceive)
		for _, es := range resp.Edges {
			if es.Source == node.ID {
				ns.Adjacency = append(ns.Adjacency, es.Target)
			}
		}
		sgb.Update = append(sgb.Update, ns)
	}
	// sgb.Reset = true
}
